/*
 * Hpolyhedron.cpp
 *
 *  Created on: Nov 20, 2012
 *      Author: romain
 */

#include <Hpolyhedron.h>
using namespace std;
using namespace arma;

namespace nltool {

Hpolyhedron::Hpolyhedron() {
	// TODO Auto-generated constructor stub

}

Hpolyhedron::~Hpolyhedron() {
	// TODO Auto-generated destructor stub
}

Hpolyhedron::Hpolyhedron(int dim, Template& mat, Vector& vec) {
	templates = mat;
	coefficients = vec;
	dimension = dim;
}

Hpolyhedron::Hpolyhedron(int dim, Template& mat, Vector& vec, Box& box) :
		templates(mat), coefficients(vec), bbox(box) {
	 dimension = dim;
}

Hpolyhedron::Hpolyhedron(Box& box) :
		bbox(box) {
	dimension = box.getDimension();
	templates = Template(2 * dimension, zeros(dimension));
	coefficients = Vector(2 * dimension);

	for (unsigned int i = 0; i < dimension; i++) {
		stdVector tmp(dimension);
		templates[i][i] = 1;
		coefficients[i] = box.getBu()[i];
		templates[i + dimension][i] = -1;
		coefficients[i + dimension] = -box.getBl()[i];
	}
}

Hpolyhedron::Hpolyhedron(Simplex& simplex) :
		bbox(Box()) {
	dimension = simplex.dim();
	vertices = simplex.getVertices();

	templates = Template(dimension + 1);
	coefficients = Vector(dimension + 1);

	for (unsigned i = 0; i < vertices.size(); i++) {
		// TODO: optimize this algorithm
		Vertices face(dimension);
		int ind = 0;
		for (unsigned j = 0; j < vertices.size(); j++) {
			if (j != i) {
				face[ind] = vertices[j];
				ind++;
			}
		}

		// construct A
		Matrix A = arma::zeros(dimension, dimension);
		for (unsigned j = 1; j < face.size(); j++) {
			for (unsigned k = 0; k < dimension; k++) {
				A(j, k) = face[j][k] - face[0][k];
			}
		}
		//		for(int k =0; k<dimension;k++)
		//			A.data[0][k] = 1;

		A(0, 0) = 1;

		// construct b
		Vector b = arma::zeros(dimension);
		b[0] = 1;

		Vector normal = arma::inv(A) * b;
		Number cst = arma::dot(normal, face[0]);

		// check if the normal is well oriented
		Number test = arma::dot(normal, vertices[i]);
		if (test > cst) {
			normal = -1 * normal;
			cst = -cst;
		}
		templates[i] = normal;
		coefficients[i] = cst;
	}
}

bool Hpolyhedron::includes(Hpolyhedron& poly) {
	assert(poly.getDimension() == dimension);
	for (unsigned i = 0; i < templates.size(); i++) {
		if (simplex(templates[i], poly.getTemplates(), poly.getCoefficients())
				> coefficients[i]){
			return false;
		}

	}
	return true;
}

bool Hpolyhedron::includes(Vector& point) {
	assert(point.size() == dimension);
	for (unsigned i = 0; i < templates.size(); i++) {
		if (dot(templates[i], point) > coefficients[i])
			return false;
	}
	return true;
}

Vector Hpolyhedron::getCoefficients() const {
	return coefficients;
}

Template Hpolyhedron::getTemplates() const {
	return templates;
}

Template Hpolyhedron::getAllTemplates() {
	Template allTemplate = templates;
	if (bbox.getBl().size() > 0) {
		Hpolyhedron tmpPoly(bbox);
		for (unsigned i = 0; i < tmpPoly.getTemplates().size(); i++)
			allTemplate.push_back(tmpPoly.getTemplates()[i]);
	}
	return allTemplate;
}

Vector Hpolyhedron::getAllCoefficients() {

	if (bbox.getBl().size() > 0) {
		Hpolyhedron tmpPoly(bbox);
		Vector allCoeff(coefficients.size() + tmpPoly.getCoefficients().size());
		for (unsigned i = 0; i < coefficients.size(); i++)
			allCoeff[i] = coefficients[i];
		for (unsigned i = 0; i < tmpPoly.getTemplates().size(); i++)
			allCoeff[i + coefficients.size()] = tmpPoly.getCoefficients()[i];
		return allCoeff;
	}
	return getCoefficients();

}

void Hpolyhedron::setCoefficients(Vector coefficients) {
	this->coefficients = coefficients;
}

vector<Hpolyhedron> Hpolyhedron::split() {
	vector<Hpolyhedron> result;

	// we use the bounding box to choose the splitting hyperplan
	Box bbox(dimension, templates, coefficients);
	stdVector bl = bbox.getBl();
	stdVector bu = bbox.getBu();
	Number maxDistance = bu[0] - bl[0];
	int maxAxis = 0;

	// find in which axis the set is the most elongated
	for (unsigned i = 1; i < dimension; i++) {
		Number distance = bu[i] - bl[i];
		if (distance > maxDistance) {
			maxDistance = distance;
			maxAxis = i;
		}
	}

	// compute the splitting constraints
	Number newCoef = bl[maxAxis] + maxDistance / 2.;
	Vector newTemplate1(zeros(dimension));
	newTemplate1[maxAxis] = 1;
	Vector newTemplate2(zeros(dimension));
	newTemplate2[maxAxis] = -1;

	// create the splitted polyhedron
	Template newTemplates1 = templates;
	Template newTemplates2 = templates;
	Vector newCoefs1 = coefficients;
	Vector newCoefs2 = coefficients;

//	newTemplates1.insert_cols(newTemplates1.n_cols,newTemplate1);
//	newTemplates2.insert_cols(newTemplates2.n_cols,newTemplate2);
	newTemplates1.push_back(newTemplate1);
	newTemplates2.push_back(newTemplate2);
	Vector tmp(1);
	tmp[0] = newCoef;
	newCoefs1.insert_rows(newCoefs1.n_elem, tmp);
	newCoefs2.insert_rows(newCoefs2.n_elem, -tmp);

//	newCoefs1.push_back(newCoef);
//	newCoefs2.push_back(-newCoef);
	Hpolyhedron poly1(dimension, newTemplates1, newCoefs1);
	Hpolyhedron poly2(dimension, newTemplates2, newCoefs2);

	// clean poly
	poly1.clean();
	poly2.clean();

	// fill the result
	result.push_back(poly1);
	result.push_back(poly2);
	return result;
}

int Hpolyhedron::clean() {
	int nbRemovedConstraints = 0;
	Template newTemplate;
	Vector newCoef;

	for (unsigned i = 0; i < templates.size(); i++) {
		// optimize in a template direction
		Number opt = simplex(templates[i], templates, coefficients);
		if (opt >= coefficients[i] - EPSILON) {
			// save the constraint
			Vector tmp(1);
			tmp[0] = coefficients[i];
			newTemplate.push_back(templates[i]);
			newCoef.insert_rows(newCoef.n_elem, tmp);
		}
	}

	templates = newTemplate;
	coefficients = newCoef;
	return nbRemovedConstraints;
}

Number Hpolyhedron::support(Vector& a) {
	return simplex(a, templates, coefficients);
}

void Hpolyhedron::inflate(Number inflat){
	for(unsigned i=0; i<coefficients.size(); i++)
		coefficients[i]+=inflat;
}

void Hpolyhedron::setTemplates(Template templates) {
	this->templates = templates;
}

Vertices Hpolyhedron::getVertices() {
	return vertices;
}

}  // namespace nltool

