/*
 * MatlabExporter.cpp
 *
 *  Created on: Nov 15, 2012
 *      Author: romain
 */

#include <MatlabExporter.h>

namespace nltool {

using namespace std;

MatlabExporter::MatlabExporter(int dim, const string fname, stdVector bl,
		stdVector bu, vector<int> projection) :
		filename(fname), uBound(bu), lBound(bl), dimension(dim), projectedAxis(
				projection), approximationTemplate(Template()) {

	assert(dim > 1);

// matlab file initialisation
	std::ofstream file(filename.data(), std::ios::out | std::ios::trunc);
	if (file) {
		file << "% AUTO-GENERATED SCRIPT" << std::endl;
		file << "% This script requires the PLOTREGION function" << std::endl;
		file << "% http://www.mathworks.com/matlabcentral/fileexchange/9261"
				<< std::endl;
		file
				<< "% Be sure to have this function in the same folder of this script"
				<< std::endl << std::endl;
		file.close();
	} else
		std::cerr << "Cannot access to " << filename << std::endl;
	cout << "projection: " << projectedAxis << endl;
}

MatlabExporter::MatlabExporter(int dim, const string fname, stdVector bl,
		stdVector bu) :
		filename(fname), uBound(bu), lBound(bl), dimension(dim), approximationTemplate(
				Template()) {

	assert(dim > 1);
	switch (dim) {
	case 1:
		projectedAxis = vector<int>(1, 0);
		break;
	case 2:
		projectedAxis = vector<int>(2, 0);
		projectedAxis[1] = 1;
		break;
	case 3:
		projectedAxis = vector<int>(3, 0);
		projectedAxis[1] = 1;
		projectedAxis[2] = 2;
		break;
	default:
		projectedAxis = vector<int>(3, 0);
		cout << "Projected dimension, enter dimensions indices in \n(0,...,"
				<< dim - 1 << ")" << endl;
		cout << " dimension: ";
		cin >> projectedAxis[0];
		cout << "\n dimension: ";
		cin >> projectedAxis[1];
		cout << "\n dimension: ";
		cin >> projectedAxis[2];
		cout << endl;
		break;
	}

// matlab file initialisation
	std::ofstream file(filename.data(), std::ios::out | std::ios::trunc);
	if (file) {
		file << "% AUTO-GENERATED SCRIPT" << std::endl;
		file << "% This script requires the PLOTREGION function" << std::endl;
		file << "% http://www.mathworks.com/matlabcentral/fileexchange/9261"
				<< std::endl;
		file
				<< "% Be sure to have this function in the same folder of this script"
				<< std::endl << std::endl;
		file.close();
	} else
		std::cerr << "Cannot access to " << filename << std::endl;
}

MatlabExporter::~MatlabExporter() {
	cout << "lBound: " << lBound << endl;
	cout << "uBound: " << uBound << endl;
	std::ofstream file(filename.data(), std::ios::out | std::ios::app);
	if (file) {



		file << "set(gca,'Xlim',[" << lBound[projectedAxis[0]] << " "
				<< uBound[projectedAxis[0]] << "]);\n";
		file << "set(gca,'Ylim',[" << lBound[projectedAxis[1]] << " "
				<< uBound[projectedAxis[1]] << "]);\n";

		file<<"Xlabel('x"<<projectedAxis[0]<<"');"<<endl;
		file<<"Ylabel('x"<<projectedAxis[1]<<"');"<<endl;

		file.close();
	} else
		std::cerr << "Cannot access to " << filename << std::endl;
}

void MatlabExporter::save(Hpolyhedron set, stdVector color) {
	Template A = set.getTemplates();
	Vector b = set.getCoefficients();
	stdVector validCoef;

	if (dimension >= 4) {
		Box bbox(set);
		Number epsi = 0.001;
		if(bbox.getMinInterval()<epsi)
			set.inflate(epsi);
		if (approximationTemplate.size() > 0) {
			A = approximationTemplate;
			b = arma::zeros(approximationTemplate.size());
			for (unsigned i = 0; i < approximationTemplate.size(); i++)
				b[i] = set.support(approximationTemplate[i]);
		} else {

			// we approximate the polyhedron by its bounding box

			stdVector bl = bbox.getBl();
			stdVector bu = bbox.getBu();

			for(unsigned i=0; i<bl.size();i++){
				if(bu[i]-bl[i]<epsi){
					bu[i]+=epsi;
					bl[i]-=epsi;
				}
			}
			Box newBbox(bl, bu);
			Hpolyhedron polyTmp(newBbox);
			A = polyTmp.getTemplates(); //
			b = polyTmp.getCoefficients();
		}
	}

	std::ofstream file(filename.data(), std::ios::out | std::ios::app);
	if (!file) {
		std::cerr << "Cannot access to " << filename << std::endl;
		return;
	}

	file << "hold on" << std::endl;
	file << std::endl << "% #########################" << std::endl;
	file << "pcolor = [" << color[0] << "," << color[1] << "," << color[2]
			<< "]; % polyhedra color" << std::endl;

	file << std::endl << "A = [ ";
	for (unsigned int i = 0; i < A.size(); ++i) {

		stdVector tmp(projectedAxis.size());
		bool valid = false;
		for (unsigned int k = 0; k < projectedAxis.size(); k++) {
			tmp[k] = A[i][projectedAxis[k]];
			if (!closeToZero(tmp[k]))
				valid = true;
		}
		if (valid) {
			if ((i > 0) && (validCoef.size() > 0))
				file << ";" << endl;
			validCoef.push_back(b[i]);
			for (unsigned k = 0; k < tmp.size(); k++) {
				file << tmp[k] << " ";
			}
		}

	}
	file << " ];" << std::endl << std::endl;

	file << "b = [";
	for (unsigned i = 0; i < validCoef.size(); ++i) {
		file << validCoef[i] << " ";
		if (i < validCoef.size() - 1)
			file << ";";
	}

	file << " ];" << std::endl << std::endl;
	file << "plotregion(-A,-b,[],[],pcolor);" << std::endl;
	file.close();
}

void MatlabExporter::save(std::vector<Hpolyhedron> sets, stdVector color) {
	for (unsigned i = 0; i < sets.size(); i++)
		save(sets[i], color);
}

void MatlabExporter::save(Simplex set) {
	std::ofstream file(filename.data(), std::ios::out | std::ios::app);
	if (!file) {
		std::cerr << "Cannot access to " << filename << std::endl;
		return;
	}
	file << "hold on" << std::endl;
	file << std::endl << "% ################ SIMPLEX" << std::endl;
	file << "x = [";
	for (unsigned i = 0; i < set.getVertices().size(); i++)
		file << set.getVertices()[i][0] << " ";
	file << set.getVertices()[0][0] << "];" << endl;
	;
	file << "y = [";
	for (unsigned i = 0; i < set.getVertices().size(); i++)
		file << set.getVertices()[i][1] << " ";
	file << set.getVertices()[0][1] << "];" << endl;
	file << "plot(x,y);" << endl;
	file.close();
}

void MatlabExporter::saveVectorField(DynamicalSystem& sys) {
	// Warning: work only with 2D system
	std::ofstream file(filename.data(), std::ios::out | std::ios::app);
	if (file) {
		double interval = (uBound[0] - lBound[0]) / 20.;
		file << "% Displaying vector field" << std::endl;
		file << "hold on" << std::endl;
		file << "x=[];\ny=[];\nu=[];\nv=[];" << std::endl;
		// displaying vector field
		double i = lBound[0];
		while (i <= uBound[0]) {
			double j = lBound[1];
			while (j <= uBound[1]) {
				Vector x(2);
				x[0] = i;
				x[1] = j;
				file << "x=[x " << i << "];" << std::endl;
				file << "y=[y " << j << "];" << std::endl;
				file << "u=[u " << sys.eval(x)[0] << "];" << std::endl;
				file << "v=[v " << sys.eval(x)[1] << "];" << std::endl
						<< std::endl;
				j += interval;
			}
			i += interval;
		}
		file << "quiver(x,y,u,v)" << std::endl;
		file.close();
	} else
		std::cerr << "Cannot access to result.m" << std::endl;
}

void MatlabExporter::findBounds(std::vector<Hpolyhedron>& sets) {
	Box tmpBox(sets[0]);
	stdVector lBoundTmp = tmpBox.getBl();
	stdVector uBoundTmp = tmpBox.getBu();
	for (unsigned i = 1; i < sets.size(); i++) {
		tmpBox = Box(sets[i]);
		for (unsigned j = 0; j < dimension; j++) {
			if (lBoundTmp[j] > tmpBox.getBl()[j])
				lBoundTmp[j] = tmpBox.getBl()[j];
			if (uBoundTmp[j] < tmpBox.getBu()[j])
				uBoundTmp[j] = tmpBox.getBu()[j];
		}
	}
	for (unsigned i = 0; i < dimension; i++) {
		Number displace = uBoundTmp[i] - lBoundTmp[i];
		lBoundTmp[i] -= 0.3 * displace;
		uBoundTmp[i] += 0.3 * displace;
	}
	lBound = lBoundTmp;
	uBound = uBoundTmp;
}

void MatlabExporter::drawTrajectory(DynamicalSystem& sys, Vector& init,
		Number maxT, Number dt) {
	// Warning: work only with 2D system
	std::ofstream file(filename.data(), std::ios::out | std::ios::app);
	if (file) {
		Vector x = init;

		file << "% Displaying trajectory" << endl;
		file << "hold on" << endl;
		file << "x=[" << init[0] << "];" << endl;
		file << "y=[" << init[1] << "];" << endl;

		Number currentT = 0;
		while (currentT < maxT) {
			currentT += dt;
			x = x + dt * sys.eval(x);
			file << "x=[x " << x[0] << "];" << std::endl;
			file << "y=[y " << x[1] << "];" << std::endl;
		}
		file << "plot(x,y);" << endl;
		file.close();
	} else
		std::cerr << "Cannot access to result.m" << std::endl;
}

}  // namespace nltool

