/*
 * OBJ format exporter for pipe flow
 */

#include <utils.h>
#include <iostream>
#include <iomanip>
#include <fstream>
#include "ObjExporter.h"

namespace nltool {


ObjExporter::ObjExporter(int dim, const string fname) :
		dimension(dim), filename(fname), verticesOffset(0),
		projectedAxis(3, 0), objectIndices(0), groupIndices(0) {
	projectedAxis[0] = 0;
	projectedAxis[1] = 1;
	projectedAxis[2] = 2;

	std::ofstream file(filename.data(), std::ios::out | std::ios::trunc);
	if (file) {
		file << "# auto generated obj"<<endl;
		file.close();
	} else
		std::cerr << "Cannot access to " << filename << std::endl;
}

ObjExporter::~ObjExporter() {
	// TODO Auto-generated destructor stub
}

void ObjExporter::save(Hpolyhedron& set) {
	std::ofstream file(filename.data(), std::ios::out | std::ios::app);
	if (file) {
		// todo: work only in 3D
		file << "o set" << objectIndices << endl;
		objectIndices++;

		int off = verticesOffset;
		if (set.getVertices().size() > 0 && dimension == 3) {
			//simplex
			Vertices vert = set.getVertices();
			file << "v " << vert[0].t();
			file << "v " << vert[1].t();
			file << "v " << vert[2].t();
			file << "v " << vert[3].t()<<endl;

			file << "f " << off + 1 << " " << off + 2 << " " << off + 3 << endl;
			file << "f " << off + 2 << " " << off + 3 << " " << off + 4 << endl;
			file << "f " << off + 3 << " " << off + 4 << " " << off + 1 << endl;
			file << "f " << off + 4 << " " << off + 1 << " " << off + 2 << endl<<endl;
			verticesOffset += 4;
		}

		else {
			Box approx(set);
			stdVector bl(approx.getBl());
			stdVector bu(approx.getBu());
			int i0 = projectedAxis[0];
			int i1 = projectedAxis[1];
			int i2 = projectedAxis[2];
			file << "v " << bl[i0] << " " << bl[i1] << " " << bl[i2] << endl;
			file << "v " << bl[i0] << " " << bl[i1] << " " << bu[i2] << endl;
			file << "v " << bl[i0] << " " << bu[i1] << " " << bl[i2] << endl;
			file << "v " << bl[i0] << " " << bu[i1] << " " << bu[i2] << endl;
			file << "v " << bu[i0] << " " << bl[i1] << " " << bl[i2] << endl;
			file << "v " << bu[i0] << " " << bl[i1] << " " << bu[i2] << endl;
			file << "v " << bu[i0] << " " << bu[i1] << " " << bl[i2] << endl;
			file << "v " << bu[i0] << " " << bu[i1] << " " << bu[i2] << endl<<endl;;

			file << "f " << off + 1 << " " << off + 3 << " " << off + 7 << " "
					<< off + 5 << endl;
			file << "f " << off + 2 << " " << off + 4 << " " << off + 8 << " "
					<< off + 6 << endl;
			file << "f " << off + 8 << " " << off + 7 << " " << off + 5 << " "
					<< off + 6 << endl;
			file << "f " << off + 4 << " " << off + 3 << " " << off + 2 << " "
					<< off + 1 << endl;
			file << "f " << off + 1 << " " << off + 2 << " " << off + 6 << " "
					<< off + 5 << endl;
			file << "f " << off + 3 << " " << off + 4 << " " << off + 8 << " "
					<< off + 7 << endl<<endl;;
			verticesOffset += 8;
		}

		file.close();
	} else
		std::cerr << "Cannot access to " << filename << std::endl;
}

void ObjExporter::save(std::vector<Hpolyhedron>& sets) {
	std::ofstream file(filename.data(), std::ios::out | std::ios::app);
		if (file) {
			// todo: work only in 3D
			file << "g group" << groupIndices << endl;
			groupIndices++;
			file.close();
		} else
			std::cerr << "Cannot access to " << filename << std::endl;
	for (unsigned i = 0; i < sets.size(); i++)
		save(sets[i]);
}


}  // namespace nltool

