/*
 * PolynomialSystem.cpp
 *
 *  Created on: Oct 14, 2010
 *      Author: romain
 */

#include <PolynomialSystem.h>
#include <iostream>
#include <sstream>
#include <cstdlib>
#include <cstring>


//#include "utils.h"
namespace nltool {


using namespace std;

// ###############################
// ## Monom functions
// ###############################

Monom::Monom(int dim, Number s, multiIndice c) :
		dimension(dim) {
	scalar = s;
	coef = c;
}

Monom::Monom(int dim, string monom, bool sign) :
		dimension(dim) {
	coef = multiIndice(dim, 0);
	scalar = 1.0;
	if (!sign)
		scalar *= -1;
	istringstream iss(monom);
	string term;
	while (getline(iss, term, '*')) {
		if ((term[0] == 'x') || (term[0] == 'X')) {
			size_t expos = term.find('^');
			if (expos == term.npos) { // no exposant
				string indice = term.substr(1, term.length());
				int i = atoi(indice.c_str());

				//mpq_set_str()
				coef[i] += 1;
			} else {
				string indice = term.substr(1, term.find('^') - 1);
				int i = atoi(indice.c_str());
				string power = term.substr(term.find('^') + 1, term.length());
				int p = atoi(power.c_str());
				coef[i] += p;
			}
		} else {
#ifdef GMP_RAT
			Number tmp;
			tmp.set_str(term.c_str(),10);
			scalar *= tmp;
#else
			scalar *= atof(term.c_str());
#endif

		}
	}
}

Monom::Monom() {
}
Monom::~Monom() {
}

Number Monom::eval(const Vector &x) {
	Number res = scalar;
	for (unsigned int i = 0; i < coef.size(); i++) {
		Number power = 1;
		for (int j = 0; j < coef[i]; j++)
			power *= x[i];
		res *= power;
	}
	return res;
}

Monom Monom::derivate(const int dx) {
	Number newScal = scalar;
	std::vector<int> newCoef = coef;
	if (newCoef[dx] <= 0)
		newScal = 0.0;
	else {
		newScal *= newCoef[dx];
		newCoef[dx] -= 1;
	}
	return Monom(dimension, newScal, newCoef);
}

// ###############################
// ## Polynom functions
// ###############################

Polynom::Polynom() :
		dimension(0), monomials(std::vector<Monom>(0)) {
}
Polynom::Polynom(int dim, string strEq) :
		dimension(dim) {

	// remove space and tabulation
	string strEq2;
	for (unsigned int i = 0; i < strEq.length(); i++)
		if ((strEq[i] != ' ') && (strEq[i] != '\t'))
			strEq2.push_back(strEq[i]);
	// cout<<"equation witout space: "<<strEq2<<endl;

	// remove the left term of the equation
	std::vector<Monom> monoms;
	string polynom = strEq2.substr(strEq2.find("=") + 1, strEq2.length());

	// recover each monoms
	istringstream iss(polynom);
	//iss<<;
	string monoms2;
	while (getline(iss, monoms2, '+')) {
		istringstream iss2(monoms2);
		// check if the first character is '-'
		bool positive = (monoms2[0] != '-');
		string monom;

		while (getline(iss2, monom, '-')) {
			if (monom.size() >= 1) {
				monoms.push_back(Monom(dim, monom, positive));
				positive = false;
			}
		}
		monomials = monoms;
	}
}

Polynom::Polynom(int dim, const std::vector<Monom> monoms) :
		dimension(dim) {
	monomials = monoms;
}

Polynom::~Polynom() {
}

void Polynom::clean() {
	std::vector<Monom> monoms;

	multiIndice multiIndiceZero(dimension, 0);
	Monom monomZero(dimension, 0, multiIndiceZero);

	for (unsigned int i = 0; i < monomials.size(); i++) {
#ifdef GMP_RAT
		if(monomials[i].scalar!=0) {
#else
		if (fabs(monomials[i].scalar) > 1e-10) {
#endif
			bool notPresent = true;
			for (unsigned int j = 0; j < monoms.size(); j++) {
				if (monomials[i].coef == monoms[j].coef) {
					notPresent = false;
					monoms[j].scalar += monomials[i].scalar;
				}
			}
			if (notPresent) {
				monoms.push_back(monomials[i]);
			}
		}
	}
	if (monoms.size() == 0)
		monoms.push_back(monomZero);
	monomials = monoms;
}

Number Polynom::eval(const Vector &x) {
	Number res = 0.0;
	for (unsigned int i = 0; i < monomials.size(); i++)
		res += monomials[i].eval(x);
	return res;
}

Polynom Polynom::derivate(const int dx) {
	std::vector<Monom> monoms;
	for (unsigned int i = 0; i < monomials.size(); i++) {
		Monom derivated = monomials[i].derivate(dx);
		monoms.push_back(derivated);
	}
	return Polynom(dimension, monoms);
}

multiIndice Polynom::computeMaxDegrees() {
	multiIndice maxDegrees(monomials[0].coef.size(), 0);
	for (unsigned int i = 0; i < monomials.size(); i++) {
		for (unsigned int j = 0; j < monomials[i].coef.size(); j++) {
			if (maxDegrees[j] < monomials[i].coef[j])
				maxDegrees[j] = monomials[i].coef[j];
		}
	}
	return maxDegrees;
}

bool Polynom::containsMultiIndice(multiIndice i, int& position) {
	for (unsigned int j = 0; j < monomials.size(); j++) {
		if (monomials[j].coef == i) {
			position = j;
			return true;
		}
	}
	return false;
}

stdVector Polynom::getBernsteinCoefficients() {
	stdVector result;
	multiIndice maxDegrees = computeMaxDegrees();
	multiIndice ind(dimension, 0);
	do {
		result.push_back(computeBernsteinCoefficient(ind, maxDegrees));
	} while (increaseMultiIndice(ind, maxDegrees));
	return result;
}

Number Polynom::computeBernsteinCoefficient(multiIndice i, multiIndice l) {
	Number coef = 0;
	multiIndice j(i.size(), 0);
	int pos;
	do {
		if (containsMultiIndice(j, pos)) {
			Number num = 1.0;
			Number den = 1.0;
			for (unsigned int k = 0; k < i.size(); k++) {
				num *= (binomialCoef(i[k], monomials[pos].coef[k]));
				den *= (binomialCoef(l[k], monomials[pos].coef[k]));
			}
			coef = coef + ((num / den) * monomials[pos].scalar);
		}
	} while (increaseMultiIndice(j, i));
	return coef;
}

//void Polynom::computeAffinebound(stdVector& al, Number& bl, stdVector& au, Number& bu){
//	multiIndice dmax = computeMaxDegrees();
//	multiIndice ind(dimension,0);
//	Number min,max;
//	stdVector BernsteinCoef;
//	do
//		BernsteinCoef.push_back(computeBernsteinCoefficient(ind,dmax));
//	while(increaseMultiIndice(ind,dmax));
//
//	// quick version: take only min max coef
//	min = BernsteinCoef[0];
//	max = BernsteinCoef[0];
//	for(unsigned i=1; i<BernsteinCoef.size(); i++){
//		if(min>BernsteinCoef[0]) min = BernsteinCoef[0];
//		if(max<BernsteinCoef[0]) max = BernsteinCoef[0];
//	}
//
//	au = max; al = min;
//	al = bl = arma::zeros(dimension);
//
//}

Polynom Polynom::replaceVariable(std::vector<Polynom> expression) {
	std::vector<Monom> monoms;
	Polynom result;
	result.dimension = dimension;
	for (unsigned int i = 0; i < monomials.size(); i++) {
		Polynom P;
		for (unsigned j = 0; j < monomials[i].coef.size(); j++) {

			if (monomials[i].coef[j] > 0 && P.monomials.size() == 0) {
				P = expression[j];
				for (int k = 1; k < monomials[i].coef[j]; k++) {
					P = P * expression[j];
				}
			} else if (monomials[i].coef[j] > 0) {
				for (int k = 0; k < monomials[i].coef[j]; k++) {
					P = P * expression[j];
				}
			}
			//P.clean();
		}
		if (P.monomials.size() == 0) {
			P.monomials.push_back(monomials[i]);
			result = result + P;
		} else
			result = result + (monomials[i].scalar * P);
		result.clean();
	}
	return result;
}

// ###############################
// ## Polynomial system functions
// ###############################

PolynomialSystem::PolynomialSystem(int dim, std::vector<Polynom> polys) :
		polynomials(polys) {
	dimension = dim;
	domainU = stdVector(dim, -100);
	domainL = stdVector(dim, 100);
	systemType = findSystemType();
	systemTimeType = CONTINUOUS;

}

PolynomialSystem::PolynomialSystem(int dim, string polysys) {
	istringstream iss(polysys);
	string eq;
	dimension = dim;
	for (int i = 0; i < dim; i++) {
		getline(iss, eq, '\n');
		polynomials.push_back(Polynom(dim, eq));
	}
	systemType = findSystemType();
	systemTimeType = CONTINUOUS;
}

PolynomialSystem::PolynomialSystem() {
}

PolynomialSystem::PolynomialSystem(int dim, int deg, int nbMon,
		const bool stable, int s) {
	dimension = dim;
	degree = deg;
	nbMonom = nbMon;

	int seed = time(NULL);
	if (s < 0)
		srand(seed);
	else
		srand(s);

	// for each dimension
	for (unsigned i = 0; i < dimension; i++) {
		std::vector<Monom> monoms;
		if (nbMon) {

			// we create a polynome of nbMon monomials
			for (int j = 0; j < nbMon; j++) {
				Number scal = 0.0;
#ifdef GMP_RAT
				while(scal==0)
				scal = Number(rand()%100+1)/50.0 -1.0;
#else
				while (fabs(scal) < 0.001)
					scal = Number(rand() % 100 + 1) / 50.0 - 1.0;
#endif
				std::vector<int> coef(dimension, 0);
				for (unsigned k = 0; k < dimension; k++)
					coef[k] = rand() % (degree + 1);

				Monom momo(dimension, scal, coef);
				monoms.push_back(momo);
			}
		} else {
			multiIndice ind(dim, 0);
			multiIndice dmax(dim, deg);
			do {
				Number scal = 0.0;

#ifdef GMP_RAT
				while(scal==0)
				scal = Number(rand()%100+1)/50.0 -1.0;
#else
				while (fabs(scal) < 0.001)
					scal = Number(rand() % 100 + 1) / 50.0 - 1.0;
#endif

				Monom momo(dimension, scal, ind);
				monoms.push_back(momo);
			} while (increaseMultiIndice(ind, dmax));
		}
		Polynom popo(dimension, monoms);
		popo.clean();
		polynomials.push_back(popo);
	}
	std::cout << "Random polynomial system:\n" << *this << std::endl;
	systemType = findSystemType();
	systemTimeType = CONTINUOUS;
}

PolynomialSystem::~PolynomialSystem() {
	// TODO Auto-generated destructor stub
}

Number PolynomialSystem::eval(const Vector &x, const int indice) {
	return polynomials[indice].eval(x);
}

Vector PolynomialSystem::eval(const Vector &x) {
	Vector result(dimension);
	for (unsigned i = 0; i < dimension; i++) {
		result[i] = polynomials[i].eval(x);
	}
	return result;
}


Vector PolynomialSystem::evalDiscretised(const Vector &x, Number dt) {
	Vector result(dimension);
	for (unsigned i = 0; i < dimension; i++) {
		result[i] = x[i] + dt * polynomials[i].eval(x);
	}
	return result;
}

void PolynomialSystem::linearInterpollation(Vector point, Matrix &A,
		Vector &b) {
	//Vector result(dimension);
	// TODO: complete linear interpollation
	A = arma::zeros(dimension, dimension);
	for (unsigned i = 0; i < dimension; i++) {
		for (unsigned j = 0; j < dimension; j++) {
			A(i,j) = ((polynomials[i]).derivate(j)).eval(point);
		}
	}
}

Number PolynomialSystem::curvature(const Vector &x, const int indice,
		const Vector &d1, const Vector &d2) {
	Number res = 0.0;
	for (unsigned int i = 0; i < d2.size(); i++) {
		Number sum = 0.0;
		for (unsigned int j = 0; j < d2.size(); j++) {
			sum += d1[j]
					* (((polynomials[indice]).derivate(i)).derivate(j)).eval(x);
		}
		res += d2[i] * sum;
	}
	return res;
}

PolynomialSystem PolynomialSystem::getDiscretizedSystem(Number dt) {
	std::vector<Polynom> result;
	for (unsigned int i = 0; i < polynomials.size(); i++) {
		multiIndice xi(dimension, 0);
		xi[i] = 1;
		Monom newMon(dimension, 1, xi);
		Polynom tmp = dt * polynomials[i];
		tmp.monomials.push_back(newMon);
		tmp.clean();
		result.push_back(tmp);
	}
	return PolynomialSystem(dimension, result);
}

PolynomialSystem PolynomialSystem::derivate(const int dx) {
	std::vector<Polynom> result;
	for (unsigned int i = 0; i < polynomials.size(); i++) {
		result.push_back(polynomials[i].derivate(dx));
	}
	return PolynomialSystem(dimension, result);
}

DynamicalSystem& PolynomialSystem::derivate(const unsigned dx) {
	std::vector<Polynom> result;
	for (unsigned int i = 0; i < polynomials.size(); i++) {
		result.push_back(polynomials[i].derivate(dx));
	}
	PolynomialSystem* ptr = new PolynomialSystem(dimension, result);
	return *ptr;
}

//#################################
// Bernstein part
//#################################

//Matrix  PolynomialSystem::computeBernsteinCoefficient(){
//	/*Matrix result=arma::zeros(dimension,dimension);
//	/*for(unsigned int i=0; i<multiIndices.size();i++){
//		stdVector coef(dimension,0.0);
//		for(unsigned int j=0; j<multiIndices.size();j++){
//			if(multiIndices[j]<=multiIndices[i]){
//				Number num=1.0;
//				Number den=1.0;
//				for(int k=0; k<dimension;k++){
//					num*=(binomialCoef(multiIndices[i][k],multiIndices[j][k]));
//					den*=(binomialCoef(maxDegrees[k],multiIndices[j][k]));
//				}
//				coef= coef + ((num/den) * coefficients[j]);
//			}
//		}
//		result.push_back(coef);
//	}*/
//	return result;
//}
//
//std::vector<Polynom> PolynomialSystem::computeUnitBoxMap(Bbox box) {
//	std::vector<Polynom> result;
//
//	for (unsigned i = 0; i < dimension; i++) {
//		Polynom poly;
//		multiIndice vi(dimension, 0);
//
//		Monom monom(dimension, box.bl[i], vi);
//		poly.monomials.push_back(monom);
//		monom.coef[i] = 1;
//		monom.scalar = box.bu[i] - box.bl[i];
//		poly.monomials.push_back(monom);
//		result.push_back(poly);
//	}
//	return result;
//}

multiIndice PolynomialSystem::getMaxDegree() {
	multiIndice max(dimension, 0);
	for (unsigned int i = 0; i < polynomials.size(); i++) {
		multiIndice maxPoly = polynomials[i].computeMaxDegrees();
		for (unsigned j = 0; j < dimension; j++) {
			if (max[j] < maxPoly[j])
				max[j] = maxPoly[j];
		}
	}
	return max;
}

PolynomialSystem PolynomialSystem::replaceVariable(
		std::vector<Polynom> expression) {
	std::vector<Polynom> polys;
	for (unsigned i = 0; i < dimension; i++) {
		polys.push_back(polynomials[i].replaceVariable(expression));
	}
	return PolynomialSystem(dimension, polys);
}

const char* PolynomialSystem::toString() {
	std::stringstream str;
	for (unsigned int i = 0; i < (polynomials).size(); i++) {
		str << "x" << i << "': " << (polynomials[i]) << std::endl;
	}
	string str2 = str.str();
	char* res = new char[str2.size() + 1];
	strcpy(res, str2.c_str());
	return (const char*) res;
}

Number PolynomialSystem::maximize(const unsigned indice,
		const Template& constraintsA, const Vector& constraintsB,
		Vector& pt) {
	// we compute a unit box map from the constraints

	Box bbox(dimension, constraintsA, constraintsB);
	std::vector<Polynom> unitBoxMap = bbox.computeUnitBoxMap();
	Polynom mappedPolynom = polynomials[indice].replaceVariable(unitBoxMap);
	stdVector bernsteinCoefs(mappedPolynom.getBernsteinCoefficients());

//	cout<<"maximize "<<polynomials[indice]<<endl;
//	cout<<"in "<< bbox;
//	cout<<"polynomial mapped to the unit box: "<<mappedPolynom<<endl;
//	cout<<"result approximation = "<< max(bernsteinCoefs)<<endl<<endl;
//	getchar();

	return max(bernsteinCoefs);
}

Number PolynomialSystem::minimize(const unsigned indice,
		const Template& constraintsA, const Vector& constraintsB,
		Vector& pt) {


	// we compute a unit box map from the constraints
	Box bbox(dimension, constraintsA, constraintsB);
	std::vector<Polynom> unitBoxMap = bbox.computeUnitBoxMap();
	Polynom mappedPolynom = polynomials[indice].replaceVariable(unitBoxMap);
	stdVector bernsteinCoefs(mappedPolynom.getBernsteinCoefficients());

//	cout<<"minimize "<<polynomials[indice]<<endl;
//	cout<<"in "<< bbox;
//	cout<<"polynomial mapped to the unit box: "<<mappedPolynom<<endl;
//	cout<<"result approximation = "<< min(bernsteinCoefs)<<endl<<endl;
//	getchar();

	return min(bernsteinCoefs);
}


SystemType PolynomialSystem::findSystemType() {
	bool isConstant = true;
	bool isLinear = true;
	bool isQuadratic = true;
	for (unsigned i = 0; i < dimension; i++) {
		for (unsigned j = 0; j < polynomials[i].monomials.size(); j++) {
			int deg = 0;
			for (unsigned k = 0; k < dimension; k++) {
				deg += polynomials[i].monomials[j].coef[k];
			}

			if (deg > 0)
				isConstant = false;
			if (deg > 1)
				isLinear = false;
			if (deg > 2)
				isQuadratic = false;
			if (!isConstant && !isQuadratic && !isLinear)
				return POLYNOMIAL;
		}
	}
	if (isConstant)
		return CONSTANT;
	if (isLinear)
		return LINEAR;
	return QUADRATIC;
}

}  // namespace nltool


