/*
 * ReachHybridization.cpp
 *
 *  Created on: Jan 9, 2013
 *      Author: romain
 */

#include "ReachHybridization.h"

namespace nltool {

using namespace std;

ReachHybridization::ReachHybridization(DynamicalSystem* nonLinearSystem,
		Hpolyhedron initialSet, Number error, Number dt) :
		originalSystem(nonLinearSystem), errorBound(error), discretizationTime(
				dt), approximationBuilder(nonLinearSystem->getDimension(),
				nonLinearSystem, false), curvatureDomainSize(-1) {

	approximationBuilder.setUseTransformation(false);
	approximationBuilder.setUseDisplacement(false);
	createApproximation(initialSet, currentState);
//	checkErrorBound(errorBound, currentState[0].approximation.getLinearSystem(),
//			currentState[0].domain);
}

ReachHybridization::~ReachHybridization() {
	// TODO Auto-generated destructor stub
}

void ReachHybridization::reach(int nbstep) {

	hybElementsResult.push_back(currentState);

	for (int i = 0; i < nbstep; i++) {
		cout << "############## Iteration " << i << endl;

		vector<HybridizationElement> newState;

		for (unsigned j = 0; j < currentState.size(); j++) {

			Hpolyhedron newReachableSet = currentState[j].approximation.reach(
					currentState[j].set);

			//cout << "reachable set: " << newReachableSet << endl;

			//cout << currentState[j].domain << endl;
			if (currentState[j].domain.includes(newReachableSet)) {
				// we can save this result
				currentState[j].set = newReachableSet;
				currentState[j].steps++;
				newState.push_back(currentState[j]);
				reachabilityResult.push_back(newState[j].set);
			} else {
				cout << "#### intersection detected" << endl;
//				reachabilityResult.push_back(newReachableSet);
//				return;

				bool requiresSplitting = currentState[j].steps == 0;
				createApproximation(currentState[j].set, newState,
						requiresSplitting);

			}
		}

		//cout << "nb current reachable sets: " << newState.size() << endl;
		currentState = newState;
	}
}

void ReachHybridization::createApproximation(Hpolyhedron& initialSet,
		std::vector<HybridizationElement>& result, bool requireSplitting) {

	Hpolyhedron approximationDomain;
	Box curvatureEstimationDomain;

	// try to create an initial approximation
	bool requireSplitting2;

	if (!requireSplitting) {
		LinearSystem linearApproximation(
				approximationBuilder.createApproximation(initialSet,
						approximationDomain, errorBound, requireSplitting2));

		if (!requireSplitting2) {
			//no set splitting required
			ReachLinear approximation(initialSet.getDimension(),
					linearApproximation, discretizationTime);

			HybridizationElement newElement(initialSet, approximation,
					approximationDomain);
			result.push_back(newElement);
			// save the result
			reachabilityResult.push_back(initialSet);
			domainsCreated.push_back(approximationDomain);

		}
	}
	if (requireSplitting || requireSplitting2) {
		// we have to split the set
		//cout << "set splitting required" << endl;
		vector<Hpolyhedron> splittedSet(initialSet.split());
		int i = 0;
		while (i < splittedSet.size()) {
			Hpolyhedron domainTmp;
			LinearSystem approximationTmp =
					approximationBuilder.createApproximation(splittedSet[i],
							domainTmp, errorBound, requireSplitting2);

			if (!requireSplitting2) {
				HybridizationElement newElement(splittedSet[i],
						ReachLinear(initialSet.getDimension(), approximationTmp,
								discretizationTime), domainTmp);
				result.push_back(newElement);
				// save the result
				reachabilityResult.push_back(newElement.set);
				domainsCreated.push_back(newElement.domain);
				i++;
			} else {
				// split again the initial set
				vector<Hpolyhedron> tmpSplittedSet(splittedSet[i].split());

				splittedSet[i] = tmpSplittedSet[0];
				splittedSet.push_back(tmpSplittedSet[1]);
			}
		}
	}

}

void checkErrorBound(Number errorBound, LinearSystem& approximation,
		DynamicalSystem& originalSystem, Hpolyhedron domain) {
	Box bbox(domain);
	int nbTest = 200;
	int positiveTest = 0;
	int negativeTest = 0;
	Number maxNegativeTest = 0.0;
	int passedTest = 0;
	int dim = approximation.getDimension();
	int nonIncludedPoint = 0;

	cout << "################# starting error bound check" << endl;

	while (passedTest < nbTest) {
		// create a random point
		Vector point(dim);
		for (int i = 0; i < dim; i++)
			point[i] = ((rand() % 100 + 1) / 100.0)
					* (bbox.getBu()[i] - bbox.getBl()[i]) + bbox.getBl()[i];

		// check if the point is inside the domain
		if (domain.includes(point)) {
			cout << "testPoint: " << point << endl;
			passedTest++;
			Vector evalOrigin = originalSystem.eval(point);
			Vector evalApprox = approximation.eval(point);

			Number norm = arma::norm(evalOrigin - evalApprox, "inf");
			cout << "Error bound check: " << norm - errorBound << endl;
			if (norm <= errorBound)
				positiveTest++;
			else {
				negativeTest++;
				if (maxNegativeTest < (norm - errorBound))
					maxNegativeTest = norm - errorBound;

			}
		} else
			nonIncludedPoint++;
	}

	cout << "Number of positive test: " << positiveTest << endl;
	cout << "Number of negative test: " << negativeTest << endl;
	cout << "Maximal negative distance: " << maxNegativeTest << endl;
	cout << "non included point computed: " << nonIncludedPoint << endl;
	wait();
}

}  // namespace nltool

