/*
 * ReachLinear.cpp
 *
 *  Created on: Nov 13, 2012
 *      Author: romain
 */

#include "ReachLinear.h"

namespace nltool {

using namespace std;
using namespace arma;

ReachLinear::ReachLinear() {
	// TODO Auto-generated constructor stub

}

ReachLinear::ReachLinear(const unsigned dim, const LinearSystem& linearSys,
		Number dt) :
		dimension(dim), linearSystem(linearSys), discretizationTime(dt), iteration(
				1) {
	//discretizedSystem = linearSystem.discretize(dt);
	Atransposed = arma::eye(dim, dim) + (dt * inv(linearSystem.getA()).t());
	eAdt = matrixExponential(linearSystem.getA(), discretizationTime);
	useFixedTemplate = true;
	//emATdt = matrixExponential(linearSystem.getA().t(),(-discretizationTime));
}

ReachLinear::~ReachLinear() {
	// TODO Auto-generated destructor stub
}

Hpolyhedron ReachLinear::reach(const Hpolyhedron& initialSet) {
	Hpolyhedron initSet = initialSet;
	unsigned nbConstraints = initSet.getTemplates().size();
	Template resultTemplate(nbConstraints,
			arma::zeros(linearSystem.getDimension()));
	Vector resultB = zeros(nbConstraints);
	Vector supportPoint, inputVertex;

	Template optimizationConstraints = initSet.getAllTemplates();
	Vector optimizationCoeff = initSet.getAllCoefficients();
	Box bbox = initSet.getBbox();
	Matrix expAdjoint = matrixExponential((-linearSystem.getA()).t(),
			discretizationTime);
	Set* tmpSet = linearSystem.getInputSet();

	Matrix invA = arma::inv(linearSystem.getA());

	for (unsigned i = 0; i < nbConstraints; i++) {
		// compute the direction successor
		Vector newNormal;
		if (!useFixedTemplate) {
			newNormal = expAdjoint * initSet.getTemplates()[i];
			newNormal = newNormal / arma::norm(newNormal, 2);
		} else
			newNormal = initSet.getTemplates()[i];

		Vector maxInputVert;
		linearSystem.getInputSet()->optimize(newNormal, maxInputVert, true);
		maxInputVert += linearSystem.getB();

		// compute a supporting point
		simplex(initialSet.getTemplates()[i], optimizationConstraints,
				optimizationCoeff, supportPoint);

		// compute its successor
		Vector nextSupportingPoint = eAdt * supportPoint
				+ (invA * eAdt) * maxInputVert - invA * maxInputVert;

		resultTemplate[i] = newNormal;
		resultB[i] = arma::dot(newNormal, nextSupportingPoint);
	}

	Hpolyhedron result(dimension, resultTemplate, resultB);
	return result;

}

}  // namespace nltool
