/*
 * ReachPolynomial.cpp
 *
 *  Created on: Nov 21, 2012
 *      Author: romain
 */

#include "ReachPolynomial.h"

namespace nltool {

using namespace std;
using namespace arma;

ReachPolynomial::ReachPolynomial(PolynomialSystem &poly) :
		polySystem(poly), dimension(poly.getDimension()), useNormalDynamic(
				false), useRefinement(false), dynamicBounds(false), firstIteration(
				false), interpolateCoef(false), P_inv_computed(false), debugMode(
				false) {

}

ReachPolynomial::~ReachPolynomial() {
	// TODO Auto-generated destructor stub
}

Template computeFacesOrientation(Hpolyhedron& set, PolynomialSystem system) {
	// compute the linear interpollation of the system

	Template A, result;
	Vector b;
	Matrix Atmp;
	Vector btmp;
	//int INFO;
	A = set.getTemplates();
	b = set.getCoefficients();
	Box bbox(set.getDimension(), set.getTemplates(), set.getCoefficients());

	Vector centroid = bbox.centroid();
	system.linearInterpollation(centroid, Atmp, btmp);

	for (unsigned int i = 0; i < A.size(); ++i) {
		Vector tmp = inv(Atmp).t() * A[i];
		result.push_back(tmp);
	}
	return result;
}

std::vector<Polynom> computeUnitBoxMap(Box& box) {
	std::vector<Polynom> result;

	for (unsigned int i = 0; i < box.getDimension(); i++) {
		Polynom poly;
		multiIndice vi(box.getDimension(), 0);

		Monom monom(box.getDimension(), box.getBl()[i], vi);
		poly.monomials.push_back(monom);
		monom.coef[i] = 1;
		monom.scalar = box.getBu()[i] - box.getBl()[i];
		poly.monomials.push_back(monom);
		result.push_back(poly);
	}
	return result;
}

Number evalBernsteinPolynom(multiIndice ind, multiIndice dmax,
		stdVector point) {
	Number result = 1;
	for (unsigned int i = 0; i < dmax.size(); i++) {
		result *= binomialCoef(dmax[i], ind[i]) * pow(point[i], ind[i])
				* pow(1 - point[i], dmax[i] - ind[i]);
	}
	return result;
}

std::vector<Number> interpolateBernsteinCoef(Polynom poly,
		std::vector<Polynom> newVariables, multiIndice dmax, Matrix P_inv) {
	unsigned n = newVariables.size();
	multiIndice I(n, 0);

	int nbcoef = 1;
	for (unsigned int i = 0; i < dmax.size(); i++)
		nbcoef *= dmax[i] + 1;

	Vector G = arma::zeros(nbcoef);

	// generate test point
	std::vector<stdVector> points;
	unsigned i = 0;
	do {
		Vector point = arma::conv_to<Vector>::from(I / dmax);
		//points.push_back(I/dmax);
		//points.push_back(generateRandomPoint(n));
		Vector evalNewVariable(n);
		for (unsigned k = 0; k < n; k++)
			evalNewVariable[k] = newVariables[k].eval(point);
		G[i] = poly.eval(evalNewVariable);
		i++;
	} while (increaseMultiIndice(I, dmax));

	// build the linear equation problem
	//	multiIndice J(n,0);
	//	for(unsigned int i=0; i<points.size();i++){
	//		// comput g(x) = f(Ax+b)
	//		Vector evalNewVariable(n);
	//		for(unsigned int k=0; k<n;k++)
	//			evalNewVariable[k] = newVariables[k].eval(points[i]);
	//		G[i] = poly.eval(evalNewVariable);
	//	}
	//cout<<"G: "<<G<<std::endl;

	int nbCoef = P_inv.n_cols;

	std::vector<Number> result(nbCoef, 0);
	for (int i = 0; i < nbCoef; i++) {
		for (int j = 0; j < nbCoef; j++) {
			result[i] += P_inv(j, i) * G[j];
		}
	}

	//cout<<"P_inv * G = "<<result<<std::endl;

	return result;

}

Matrix computeBernsteinCoefInterpolation(multiIndice dmax) {
	int n = dmax.size();
	multiIndice I(n, 0);

	// generate test point
	std::vector<stdVector> points;
	do {
		points.push_back(I / dmax);
		//points.push_back(generateRandomPoint(n));
	} while (increaseMultiIndice(I, dmax));

	Matrix P = arma::zeros(points.size(), points.size());
	//cout<<"P: "<<P<<std::endl;

	// build the linear equation problem
	multiIndice J(n, 0);
	for (unsigned int i = 0; i < points.size(); i++) {
		// add row to P
		Vector tmp = arma::zeros(points.size());
		for (unsigned j = 0; j < points.size(); j++)
			P(i, j) = evalBernsteinPolynom(J, dmax, points[j]);
		increaseMultiIndice(J, dmax);
	}

	//	std::cout<<"linear equation generation: "<<clock()/(Number)CLOCKS_PER_SEC-startTime<<" s"<<std::endl;
	//	startTime = clock()/(Number)CLOCKS_PER_SEC;

	// solve the problem
	Matrix P_inv = arma::inv(P);
	return P_inv;

}

void performSetVariableChange(Vector & a, Number & b,
		std::vector<Polynom>& newVar) {

	int dim = a.size();
	int pos;
	Vector aNew(zeros(dim));
	Number bNew = b;
	for (int i = 0; i < dim; i++) {
		if (a[i] != 0) {
			for (int j = 0; j < dim; j++) {
				multiIndice multi(dim, 0);
				multi[j] = 1;

				if (newVar[i].containsMultiIndice(multi, pos)) {
					aNew[j] += newVar[i].monomials[pos].scalar * a[i];
				}
			}
			multiIndice multi(dim, 0);
			if (newVar[i].containsMultiIndice(multi, pos)) {
				bNew += -newVar[i].monomials[pos].scalar * a[i];
			}
		}
	}
	a = aNew;
	b = bNew;
}

void testVariableChange() {
	using namespace std;

	Vector a(ones(2));
	Number b = 2;
	std::vector<Polynom> unitBoxMap;

	Polynom q1(2, "x0'=x0 + x1 + 2");
	std::cout << "q1: " << q1 << endl;
	Polynom q2(2, "x1'=-x0 + 2*x1 + 3");
	std::cout << "q2: " << q2 << endl;
	unitBoxMap.push_back(q1);
	unitBoxMap.push_back(q2);

	performSetVariableChange(a, b, unitBoxMap);
	std::cout << "a: " << a << "<=" << b << std::endl;
}

void performSetVariableChange(Template& A, Vector& b,
		std::vector<Polynom>& newVar) {
	// by convention each poly have the following form:
	// x_i = scalar + scalar*z_i
	for (unsigned int i = 0; i < A.size(); i++) {
		performSetVariableChange(A[i], b[i], newVar);
	}
}

int computeDmax(PolynomialSystem polys) {
	int max = 0;
	for (unsigned int i = 0; i < polys.polynomials.size(); i++) {
		int maxTmp = 0;
		for (unsigned int j = 0; j < polys.polynomials[i].monomials.size();
				j++) {
			maxTmp = sum(polys.polynomials[i].monomials[j].coef);
			if (maxTmp > max)
				max = maxTmp;
		}
	}
	return max;
}

std::vector<Polynom> computeUnitBoxMap(int dim, Template Ctild, stdVector d) {
	std::vector<Polynom> result;
	int nbBound = d.size() / 2;
	// build problem parameter:
	Matrix D = arma::zeros(nbBound, nbBound);
	Matrix C = arma::zeros(nbBound, nbBound);

	for (int i = 0; i < nbBound; i++) {
		D(i, i) = d[i] + d[nbBound + i];
		for (int j = 0; j < nbBound; j++)
			C(i, j) = Ctild[i][j];
	}

	// compute inverse
	Matrix C_inv = arma::inv(C);
	Matrix Ci_D = C_inv * D;

	//std::cout<<"Ci_D: "<<std::endl<<Ci_Di<<std::endl;

	// create the polynomial vector
	for (int i = 0; i < dim; i++) {
		std::vector<Monom> monoms;
		for (int j = 0; j < dim; j++) {
			multiIndice multi(dim, 0);
			multi[j] = 1;
			monoms.push_back(Monom(dim, Ci_D(i, j), multi));
		}
		multiIndice multi(dim, 0);
		Number scalar = 0;
		for (int j = 0; j < nbBound; j++)
			scalar += C_inv(i, j) * (-d[nbBound + j]);
		monoms.push_back(Monom(dim, scalar, multi));
		result.push_back(Polynom(dim, monoms));
	}

	return result;
}

int ReachPolynomial::computedPinv(PolynomialSystem discret) {
	multiIndice maxDegrees;
	if (dynamicBounds)
		maxDegrees = multiIndice(discret.getDimension(), computeDmax(discret));
	else
		maxDegrees = discret.getMaxDegree();
	P_inv = computeBernsteinCoefInterpolation(maxDegrees);
	P_inv_computed = true;
	return P_inv.n_cols;
}

int ReachPolynomial::computeBernsteinCoef(const int dim, Polynom& poly,
		multiIndice& maxDegrees, std::vector<Number>& coef,
		const std::vector<Polynom> unitBoxMap) {

	if (!interpolateCoef) {
		// perform variable change
		Polynom polyUnitBox = poly.replaceVariable(unitBoxMap);
		//maxDegrees = polyUnitBox.computeMaxDegrees();
		multiIndice ind(dim, 0);
		do
			coef.push_back(
					polyUnitBox.computeBernsteinCoefficient(ind, maxDegrees));
		while (increaseMultiIndice(ind, maxDegrees));
	} else {
		coef = interpolateBernsteinCoef(poly, unitBoxMap, maxDegrees, P_inv);
	}

	// compute the real number of bernstein coefficient <= maxDegree
	int nbCoef = 1;
	for (unsigned int j = 0; j < maxDegrees.size(); j++)
		nbCoef *= maxDegrees[j] + 1;
	return nbCoef;
}

Hpolyhedron ReachPolynomial::reach(Hpolyhedron &initialSet,
		Template& templates) {
	// ################################### INITIALIZATION STEP
	unsigned dim = polySystem.getDimension();

	// normalise initial set constraints
	unsigned nbConstraintsTmp = initialSet.getTemplates().size();
	unsigned nbInitialConstraints = initialSet.getTemplates().size() + 2 * dim;

	Template initialSetA(nbInitialConstraints);
	Vector initialSetb(nbInitialConstraints);
	for (unsigned int i = 0; i < nbConstraintsTmp; i++) {
		initialSetb[i] = initialSet.getCoefficients()[i];
		initialSetA[i] = initialSet.getTemplates()[i];
	}

	// compute unit box map
	std::vector<Polynom> unitBoxMap;
	Box bbox(initialSet.getDimension(), initialSet.getTemplates(),
			initialSet.getCoefficients());
	unitBoxMap = computeUnitBoxMap(bbox);

	// adding bounding box constraints:
	Template ABox;
	Vector bBox;
	Hpolyhedron boxPoly(bbox);
	ABox = boxPoly.getTemplates();
	bBox = boxPoly.getCoefficients();

	for (unsigned int j = 0; j < ABox.size(); j++) {
		initialSetA[j + nbConstraintsTmp] = ABox[j];
		initialSetb[j + nbConstraintsTmp] = bBox[j];
		templates.push_back(ABox[j]);
	}
	//	}

	// we have to make a change of variable to work only on the unit box

	performSetVariableChange(initialSetA, initialSetb, unitBoxMap);

	Vector bResult(templates.size());	// result vector

	// compute dmax
	multiIndice maxDegrees = polySystem.getMaxDegree();

	// compute P-1 (if necessary)

	if (interpolateCoef) {
		if (!P_inv_computed) {
			P_inv = computeBernsteinCoefInterpolation(maxDegrees);
			P_inv_computed = true;
		}
	}

	// ################################### ITERATION STEP

	// for each template
	for (unsigned int i = 0; i < templates.size(); i++) {

		// compute hi.p(x)
		Polynom polyTemplate = templates[i] * polySystem;

		// compute bernstein coefficient
		std::vector<Number> BernsteinCoef;
		int nbCoef = computeBernsteinCoef(dim, polyTemplate, maxDegrees,
				BernsteinCoef, unitBoxMap);

		// ########### LP Constraints creation

		// objective (bernstein coefficient)
		int nbYi = BernsteinCoef.size();

		// inequalities:
		stdMatrix ineqA;
		stdVector ineqB;

		for (unsigned int j = 0; j < initialSetA.size(); j++) {
			multiIndice ind2(polySystem.getDimension(), 0);
			stdVector ineq;

			do {
				Vector tmp = arma::conv_to<Vector>::from(ind2 / maxDegrees);
				ineq.push_back(dot(initialSetA[j], tmp));
			} while (increaseMultiIndice(ind2, maxDegrees));

			if (notNul(ineq)) {
				ineqA.push_back(ineq);
				ineqB.push_back(initialSetb[j]);
			}
		}

		int ind3 = 0;
		multiIndice ind2(polySystem.getDimension(), 0);
		do {
			stdVector ineq(nbYi, 0);
			ineq[ind3] = 1;
			ineqA.push_back(ineq);
			ineqB.push_back(
					evalBernsteinPolynom(ind2, maxDegrees, ind2 / maxDegrees));
			ind3++;
		} while (increaseMultiIndice(ind2, maxDegrees));

		// equalities;
		stdMatrix eqA(1, stdVector(nbCoef, 1.0));
		stdVector eqB(1, 1.0);

		// domain
		stdVector unitU(nbCoef, 1.0);
		stdVector unitL(nbCoef, 0.0);

		// SOLVING!!
		int err = 0;
		Number result = simplex(BernsteinCoef, ineqA, ineqB, eqA, eqB, unitL,
				unitU, err);

		if (err) {
			//cerr <<"LP solver failed\n";
			std::cout
					<< "### ERROR: LP solver failed !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!"
					<< std::endl << std::endl;
		} else {
			if (debugMode)
				std::cout << "LP optimum: " << result << std::endl << std::endl;
		}
		// save result
		bResult[i] = result;
	} // end for each template

	// ################################### RESULT EXTRACTION STEP

	// extract bounding box
	Vector nextB(nbConstraintsTmp);
	Template nextA(nbConstraintsTmp);
	int indiceStart = bResult.size() - 2 * dim;
	for (unsigned int i = 0; i < nbConstraintsTmp; i++) {
		nextB[i] = bResult[i];
		nextA[i] = templates[i];
	}

	Box newBbox(dim, templates, bResult);
	Hpolyhedron res(initialSet.getDimension(), nextA, nextB, newBbox);
	return res;

}

Hpolyhedron ReachPolynomial::reach(Hpolyhedron &initialSet) {

	// ################################### INITIALIZATION STEP
	Template initialSetA = initialSet.getTemplates();

	// we compute the template
	Template templates;
	if (useNormalDynamic)
		templates = computeFacesOrientation(initialSet, polySystem);
	else
		templates = initialSetA; // no computation

	return reach(initialSet, templates);
}

bool ReachPolynomial::isUseNormalDynamic() const {
	return useNormalDynamic;
}

void ReachPolynomial::setUseNormalDynamic(bool useNormalDynamic) {
	this->useNormalDynamic = useNormalDynamic;
}

}  // namespace nltool
