/*
 * Simplex.cpp
 *
 *  Created on: Nov 21, 2012
 *      Author: romain
 */

#include "Simplex.h"

namespace nltool {


using namespace std;
using namespace arma;

bool Simplex::isRegular(){
	double epsi = 1e-9;
	double size = -1;
	for(unsigned i=0; i<nb_vertices(); i++){
		for(unsigned j=0; j<nb_vertices() ; j++){
			if(i!=j){
				double n = 0;
				for(unsigned d=0; d<dimension ; d++){
					n += (verticesData[i][d]-verticesData[j][d]) * (verticesData[i][d]-verticesData[j][d]);
				}
				if(size<0)
					size=sqrt(n);
				else
					if(fabs(size-n)>epsi)
						return false;
			}
		}
	}
	return true;
}


Simplex::~Simplex() {
	// TODO Auto-generated destructor stub
}
Simplex::Simplex(int dim, Vertices& vert):verticesData(vert){
	dimension = dim;
}
Simplex::Simplex():verticesData(Vertices(0)){
	dimension = 0;
}

int Simplex::dim(){
	return dimension;
}

Vertices Simplex::getVertices(){
	return verticesData;
}

unsigned int Simplex::nb_vertices(){return verticesData.size();}

Vertices Simplex::vertices(){return getVertices();}

Vector Simplex::centroid(){
	Vector center = zeros(dimension);
	for(unsigned i=0;i<verticesData.size();i++){
		center += verticesData[i];
	}
	return center/verticesData.size();
}

void Simplex::linearInterpolation(DynamicalSystem& system, Matrix &A, Vector &b) {
	A = zeros(dimension, dimension);
	b = Vector(dimension);
	Matrix matA = zeros((dimension+1)*dimension,(dimension+1)*dimension);
	Matrix matB = zeros((dimension+1)*dimension);
	Vertices vertices(getVertices());

	for(unsigned i=0; i<vertices.size(); i++){
		for(unsigned j=0; j<dimension; j++){
			for(unsigned k=0; k<dimension; k++)
				matA(i*dimension + j,j*dimension+k) = vertices[i][k];
			matA(i*dimension + j, dimension*dimension+j)=1;
			matB(i*dimension + j) = system.eval(vertices[i])[j];
		}
	}

	Matrix matRes = solve(matA, matB);

	for(unsigned i=0; i<dimension; i++)
		for(unsigned j=0; j<dimension; j++)
			A(i,j) = matRes[i*dimension+j];
	for(unsigned i=0; i<dimension; i++)
		b[i] = matRes[dimension*dimension + i];

}


}  // namespace nltool

