/*
 * BallSet.h
 *
 *  Created on: Nov 13, 2012
 *      Author: romain
 */

#ifndef BALLSET_H_
#define BALLSET_H_
#include "utils.h"
#include "Set.h"

namespace nltool {

class Ball: public Set {
public:
	/**
	 * Non-consistent constructor
	 */
	Ball();

	/**
	 * Creates a ball representation
	 * @param dim dimension of the set
	 * @param rad radius of the ball
	 * @param center center of the ball
	 */
	Ball(const unsigned dim, const Number rad, const stdVector center);

	/**
	 * Creates a ball representation centered at the origin
	 * @param dim dimension of the set
	 * @param rad radius of the ball
	 */
	Ball(const unsigned dim, const Number rad = 0);

	/**
	 * Destructor
	 */
	virtual ~Ball();

	/**
	 * Performs a quick linear optimisation in this ball
	 * @param direction optimization
	 * @param optPoint the problem solution
	 * @param max is a maximization problem (otherwise minimization)
	 * @return optimization result
	 */
	virtual Number optimize(const Vector & direction, Vector & optPoint,
			bool max = true);

	/*
	 * ##########################################
	 * GETTTERS and SETTERS
	 * ##########################################
	 */

	stdVector getCenter() const;
	Number getRadius() const;
	void setCenter(stdVector center);
	void setRadius(Number radius);
private:
	stdVector center;
	Number radius;
};

inline std::ostream& operator<<(std::ostream& out, Ball ball) {
	out << "Ball radius: " << ball.getRadius() << " ball center: "
			<< ball.getCenter() << std::endl;
	return out;
}

} // namespace nltool

#endif /* BALLSET_H_ */
