/*
 * DynamicalSystem.h
 *
 *  Created on: Nov 13, 2012
 *      Author: romain
 */

#ifndef CONTINUOUSSYSTEM_H_
#define CONTINUOUSSYSTEM_H_

#include "Set.h"
#include "utils.h"

namespace nltool {


enum SystemType {
	CONSTANT, LINEAR, QUADRATIC, POLYNOMIAL, OTHER, POINTER
};
enum SystemTimeType {
	CONTINUOUS, DISCRETE
};

/**
 * Interface class for all dynamical system
 */
class DynamicalSystem {
public:

	/**
	 * Non-consistent constructor
	 */
	DynamicalSystem();

	/**
	 * Destructor
	 */
	virtual ~DynamicalSystem();

	/**
	 * Evaluation function
	 * @param x a state-space point
	 * @return evaluation results for each variable dynamics
	 */
	virtual Vector eval(const Vector& x);

	/**
	 * Derivation function (only supported by polynomial system)
	 * @param dx derivation variable
	 * @return the derivated dynamical systems
	 */
	virtual DynamicalSystem& derivate(const unsigned dx);

	/**
	 * Displays the type of the system
	 */
	void printType();

	/*
	 * ##########################################
	 * GETTTERS and SETTERS
	 * ##########################################
	 */

	unsigned getDimension() const;
	void setDimension(unsigned dimension);

	SystemTimeType getSystemTimeType() const;
	void setSystemTimeType(SystemTimeType systemTimeType);
	SystemType getSystemType() const;


protected:
	unsigned dimension; // system dimension
	SystemType systemType; // system type
	SystemTimeType systemTimeType; // discret of continuous time
	Number discretizedTime; // discretization time in case of discrete time type

};

}  // namespace nltool


#endif /* CONTINUOUSSYSTEM_H_ */
