/*
 * LinearSystem.h
 *
 *  Created on: Nov 13, 2012
 *      Author: romain
 */

#ifndef LINEARSYSTEM_H_
#define LINEARSYSTEM_H_

#include "DynamicalSystem.h"
#include "Set.h"

namespace nltool {

class LinearSystem: public DynamicalSystem {
public:
	/**
	 * Non-consistent constructor
	 */
	LinearSystem();

	/**
	 * Constructor or an affine system without input
	 * @param dim the dimension of the system
	 * @param C a matrix representing the linear part of the affine dynamic
	 * x'=Cx+d
	 * @param d a vector representing a constant part of the affine dynamic
	 * x'=Cx+d
	 */
	LinearSystem(const unsigned dim, const Matrix C, const Vector d);

	/**
	 * Constructor or an affine system with input
	 * @param dim the dimension of the system
	 * @param C a matrix representing the linear part of the affine dynamic
	 * x'=Cx+d
	 * @param d a vector representing a constant part of the affine dynamic
	 * x'=Cx+d
	 * @param input the input set
	 */
	LinearSystem(const unsigned dim, const Matrix C, const Vector d,
			Set* input);

	/**
	 * Destructor
	 */
	virtual ~LinearSystem();

	/**
	 * Evaluation function
	 * @param x the state of the system
	 * @return
	 */
	Vector eval(const Vector& x);

	//LinearSystem& operator=(const LinearSystem&);

	/*
	 * ##########################################
	 * GETTTERS and SETTERS
	 * ##########################################
	 */

	Matrix getA() const;
	Vector getB() const;
	Set* getInputSet() const;

	void setA(Matrix a);
	void setB(Vector b);
	void setInputSet(Set* inputSet);

private:
	Matrix A;
	Vector b;
	Set* inputSet;
};

inline std::ostream& operator<<(std::ostream& out, LinearSystem linearSys) {
	out << "A: \n" << linearSys.getA() << std::endl << "b: \n"
			<< linearSys.getB() << std::endl;
	return out;
}

}

#endif /* LINEARSYSTEM_H_ */
