/*
 * MatlabExporter.h
 *
 *  Created on: Nov 15, 2012
 *      Author: romain
 */

#ifndef MATLABEXPORTER_H_
#define MATLABEXPORTER_H_

#include "PolynomialSystem.h"
#include "Hpolyhedron.h"
#include "iostream"
#include "string"
#include "fstream"
#include "iomanip"
#include "stdio.h"

namespace nltool {

class MatlabExporter {
public:
	MatlabExporter(int dim, const string fname = "noname.m", stdVector bl =
			stdVector(2, -2), stdVector bu = stdVector(2, 2));
	MatlabExporter(int dim, const string fname, stdVector bl, stdVector bu,
			vector<int> projection);
	virtual ~MatlabExporter();

	void save(Hpolyhedron set, stdVector color = stdVector(3, 0.6));
	void save(std::vector<Hpolyhedron> sets,
			stdVector color = stdVector(3, 0.6));
	void save(Simplex set);
	void saveVectorField(DynamicalSystem& sys);
	void drawTrajectory(DynamicalSystem& sys, Vector& init, Number maxT,
			Number dt = 0.01);

	void findBounds(std::vector<Hpolyhedron>& sets);

	void setApproximationTemplate(const Template& approximationTemplate) {
		this->approximationTemplate = approximationTemplate;
	}

private:
	Template approximationTemplate;
	string filename;
	vector<int> projectedAxis;
	stdVector lBound;
	stdVector uBound;
	int displayEvery;
	int dimension;
};

}  // namespace nltool

#endif /* MATLABEXPORTER_H_ */
