/*
 * ReachHybridization.h
 *
 *  Created on: Jan 9, 2013
 *      Author: romain
 */

#ifndef REACHHYBRIDIZATION_H_
#define REACHHYBRIDIZATION_H_

#include "utils.h"
#include "ReachLinear.h"
#include "Hpolyhedron.h"
#include "ApproximationBuilder.h"

namespace nltool {

class ReachHybridization {
public:

	struct HybridizationElement {
		HybridizationElement(Hpolyhedron s, ReachLinear a, Hpolyhedron d) :
				set(s), approximation(a), domain(d), steps(0) {
		}
		Hpolyhedron set;
		ReachLinear approximation;
		Hpolyhedron domain;
		int steps;
	};

	/**
	 * Constructor
	 * @param nonLinearSystem a non-linear system to analyse
	 * @param initialSet the initial set containing the possible initial states
	 * @param errorBound a desired precision for the approximation
	 * @param dt a discrertization time step
	 */
	ReachHybridization(DynamicalSystem* nonLinearSystem, Hpolyhedron initialSet,
			Number error, Number dt);

	/**
	 * Destructor
	 */
	virtual ~ReachHybridization();

	/**
	 * Reachability function which compute the reachable
	 * @param nbstep the number of reachable set computed
	 */
	void reach(int nbstep = 1);

	// Getters and Setters

	Number getDiscretizationTime() const {
		return discretizationTime;
	}

	void setDiscretizationTime(Number discretizationTime) {
		this->discretizationTime = discretizationTime;
	}

	Number getErrorBound() const {
		return errorBound;
	}

	void setErrorBound(Number errorBound) {
		this->errorBound = errorBound;
	}

	std::vector<Hpolyhedron> getReachabilityResult() const {
		return reachabilityResult;
	}

	std::vector<Hpolyhedron> getDomainsCreated() const {
		return domainsCreated;
	}

	std::vector<std::vector<HybridizationElement> > getHybElementsResult() const {
		return hybElementsResult;
	}

private:

	void createApproximation(Hpolyhedron& initialSet,
			std::vector<HybridizationElement>& result, bool requireSplitting =
					false);

	std::vector<HybridizationElement> currentState;
	DynamicalSystem* originalSystem;

	ApproximationBuilder approximationBuilder;
	Number errorBound;
	Number discretizationTime;
	Number curvatureDomainSize;

	std::vector<std::vector<HybridizationElement> > hybElementsResult;
	std::vector<Hpolyhedron> reachabilityResult;
	std::vector<Hpolyhedron> domainsCreated;
};

void checkErrorBound(Number errorBound, LinearSystem& approximation, DynamicalSystem& originalSystem,
		Hpolyhedron domain);

}  // namespace nltool

#endif /* REACHHYBRIDIZATION_H_ */
